#!/usr/bin/perl -w
use strict;
use Parse::RecDescent;
use File::Basename;

# Prerequisites:
#
#  sudo apt-get install make cpanminus        # (makes it much easier to install Perl modules)
#  sudo cpanm install Parse::RecDescent



#$::RD_WARN  = 1;
#$::RD_TRACE  = 1; # Show Parse::RecDescent operation - useful for debugging

# The following array of patterns is used to determine what properties to
# comment. They are applied in order, and the flag value of the last matching
# pattern wins.
my @patterns = (

# comment flag (0==no, 1==yes, 2==print)
# |   base filename
# |   |       selector pattern
# |   |       |        property pattern
# |   |       |        |
 [1,  qr{},   qr{(title|section|pre)},   qr{^(margin|padding|font)}],
 [1,  qr{},   qr{(page)},   qr{^(margin|padding)}],
 [1,  qr{^p-pages-and-headers},   qr{},   qr{^content}],
 [1,  qr{^(p-numbering|p-i18n).*},   qr{},   qr{^content}],
# [1,  qr{},   qr{},   qr{^margin|padding}],
# [1,  qr{},   qr{},   qr{^(font|margin|padding|background|text-align|text-decoration|border|content|color|background)}],
#
# [0,  qr{^-subject-scheme-map\.css$},      qr{},   qr{}],
# [0,  qr{^p-side-notes\.css$},             qr{},   qr{}],
# [0,  qr{^-page\.css$},                    qr{},   qr{}],
# [0,  qr{^-topic-legacy-conversion\.css$}, qr{},   qr{}],
# [0,  qr{^-topic-metadata-prolog\.css$},   qr{},   qr{}],

# next if $filename =~ m!-topic-body-lists\.css$!;
# my $origfile = my $file = readfile($filename);
# $file =~ s!^([ \t]+margin(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;
# $file =~ s!^([ \t]+padding(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;
# $file =~ s!^([ \t]+text-align(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;
# $file =~ s!^([ \t]+text-decoration(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;
# $file =~ s!^([ \t]+border(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;
# $file =~ s!^([ \t]+font(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;
# $file =~ s!^([ \t]+content(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;
# $file =~ s!^([ \t]+color(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;
# $file =~ s!^([ \t]+background(-[\w\-]+)?:.*;) *(\r?)$!/*$1 - MODIFIED*/$3!mg;

);


# The following subroutine is called on-the-fly during CSS parsing to return
# a commented or uncommented property declaration. The $basename, $selector,
# and $property variables are defined globally so they can be shared
# throughout the code.
our $basename;
our %matches = ();
sub add_comment {
 my $comment_flag;
 my ($text, $selector, $property, $value) = @_;
 foreach my $patternref (@patterns) {
  my ($flag, $basepat, $selpat, $proppat) = @$patternref;
  $comment_flag = $flag if $::basename =~ $basepat && $selector =~ $selpat && $property =~ $proppat;
 }
 
 $::matches{::trim($::basename)}->{::trim($selector)}->{::trim("$property: $value")} = 1 if ($comment_flag or 0) == 2;
 return $text if !defined($comment_flag) || $comment_flag == 2;
 return "/* $text - COMMENTED */" if $comment_flag == 1;
 return "$text /* KEPT */" if $comment_flag == 0;
}
sub trim { (my $s = shift) =~ s!(^\s+|\s+$)!!ss; $s =~ s!\s\s+! !gs; return $s; }


# These subroutines take Parse::RecDescent's array-reference structures
# and flatten them to text strings. We use this as the default action so
# that we stream the actual content through parsing into the return result.
sub flatten { map { ref $_ eq 'ARRAY' ? flatten(@$_) : $_ } @_ }
sub txt { return join('', flatten(@_[1..$#_])); }


# The following grammar is a simplified version of the grammar described at
#
#   TO-DO
#
# and implemented by CSS::Parse::Heavy:
#
#   (https://metacpan.org/pod/CSS::Parse::Heavy)
#
# It has also been modified to accept additional Oxygen XML CSS constructs.
my $grammar = <<'EOS';
<skip:"">
<autoaction:{$return = ::txt(@item); 1;}>
  stylesheet:		(WS | statement)(s)
  statement:		atrule | ruleset
  atrule:		<rulevar: local $selector>
  atrule:		ATKEYWORD OWS any(s?) {$selector = ::txt(@item); '';} (atblock | ';' )
  atblock:		'{' OWS (ruleset | declaration | ';' | WS)(s?) '}'
  ruleset:		<rulevar: local $selector>
  ruleset:		OWS selector(?) '{' OWS declaration(?) (';' OWS declaration(?))(s?) '}'
  selector:		any(s)					{$selector = $return = ::txt(@item); 1;}
  declaration:		<rulevar: local $property>
  declaration:		<rulevar: local $value>
  declaration:		property OWS ':' OWS value		{$return = main::add_comment(main::txt(@item), $selector, $property, $value); 1;}
  property:		m{-?} IDENT				{$property = $return = ::txt(@item); 1;}
  value:		(any | block OWS | ATKEYWORD OWS)(s)	{$value = $return = ::txt(@item); 1;}
  block:		'{' OWS (any | block | ATKEYWORD OWS | ';' )(s?) '}'
  any:			any_item OWS
  any_item:		URI | PERCENTAGE | DIMENSION | NUMBER | STRING | HASH | UNICODERANGE | IDENT | INCLUDES | DASHMATCH | '(' any(s?) ')' | '[' any(s?) ']' | FUNCTION | DELIM

  ##
  ## TOKENS
  ##
  IDENT:		macro_ident
  ATKEYWORD:		'@' macro_ident
  STRING:		macro_string
  HASH:			'#' macro_name
  NUMBER:		macro_num
  PERCENTAGE:		macro_num '%'
  DIMENSION:		macro_num macro_ident
  URI:			'url(' macro_w macro_string macro_w ')' | 'url(' macro_w (/[!#$%&*-~]/ | macro_nonascii | macro_escape)(s?) macro_w ')'
  UNICODERANGE:		/U\+[0-9A-F?]{1,6}(-[0-9A-F]{1,6})?/
  WS:			m{([ \t\r\n\f]|/\*.*?\*/|//.*?\r?\n)+}s
  OWS:			WS(?)
  FUNCTION:		macro_ident '('
  INCLUDES:		'~='
  DASHMATCH:		'|='
  DELIM:		/[^0-9a-zA-Z\{\}\(\)\[\];]/

  ##
  ## MACROS
  ##
  macro_ident:		macro_nmstart macro_nmchar(s?)
  macro_name:		macro_nmchar(s)
  macro_nmstart:	/[a-zA-Z]/ | macro_nonascii | macro_escape
  macro_nonascii:	/[^\0-\177]/
  macro_unicode:	/\[0-9a-f]{1,6}[ \n\r\t\f]?/
  macro_escape:		macro_unicode | /\\[ -~\200-\4177777]/
  macro_nmchar:		/[a-z0-9-]/ | macro_nonascii | macro_escape
  macro_num:		/[0-9]+|[0-9]*\.[0-9]+/
  macro_string:		macro_string1 | macro_string2
  macro_string1:	'"' (/[^\n\r\f\\"]/ | '\\' macro_nl | "'" | macro_nonascii | macro_escape )(s?) '"'
  macro_string2:	"'" (/[^\n\r\f\\']/ | '\\' macro_nl | '"' | macro_nonascii | macro_escape )(s?) "'"
  macro_nl:		/\n|\r\n|\r|\f/
  macro_w:		/[ \t\r\n\f]*/
EOS
my $parser = Parse::RecDescent->new($grammar);


# Iterate through the provided files, parse them, write them out in-place.
while (my $filename = shift) {
 $::basename = basename($filename);
 open my $fh, '<', $filename or die "can't open '$filename' for reading: $!";
 binmode($fh, ":encoding(utf-8)");
 local $/ = undef;
 my $contents = <$fh>;
 close $fh;

 if (defined (my $result = $parser->stylesheet(\$contents)) && $contents eq '') {
  open my $fh, '>', $filename or die "can't open '$filename' for writing: $!";
  binmode($fh, ":encoding(utf-8)");
  print $fh $result;
  close $fh;
  print $filename." SUCCEEDED\n";
 } else {
  print $filename." FAILED:\n";
  print "  UNPARSED:\n".$contents."\n\n";
 }
}


# Print matching constructs (flag==2), if they exist.
if (%::matches) {
 foreach my $basename (sort keys %::matches) {
  print "$basename:\n";
  foreach my $selector (sort keys %{$::matches{$basename}}) {
   print "\t$selector\n";
   foreach my $declaration (sort keys %{$::matches{$basename}->{$selector}}) {
    print "\t\t$declaration\n";
   }
  }
 } 
}

